import json
from dataclasses import dataclass
from functools import lru_cache, partial
from typing import Any, Dict, List, Union, Optional

import html2text
import openai
import pandas as pd
import requests
import writer
from langchain_community.embeddings.huggingface import HuggingFaceEmbeddings
from langchain_community.llms import Writer
from langchain_community.document_loaders import DataFrameLoader
from langchain_community.vectorstores import FAISS, Chroma
from pydantic import BaseModel, Extra, Field, field_validator, ValidationInfo


from mindsdb.integrations.handlers.chromadb_handler.chromadb_handler import get_chromadb
from mindsdb.integrations.handlers.rag_handler.exceptions import (
    InvalidOpenAIModel,
    InvalidPromptTemplate,
    InvalidWriterModel,
    UnsupportedLLM,
    UnsupportedVectorStore,
)
from langchain_core.callbacks import StreamingStdOutCallbackHandler
from langchain_core.documents import Document
from langchain_core.embeddings import Embeddings
from langchain_core.vectorstores import VectorStore

DEFAULT_EMBEDDINGS_MODEL = "BAAI/bge-base-en"

SUPPORTED_VECTOR_STORES = ("chromadb", "faiss")

SUPPORTED_LLMS = ("writer", "openai")

## Default parameters for RAG Handler

# this is the default prompt template for qa
DEFAULT_QA_PROMPT_TEMPLATE = """
Use the following pieces of context to answer the question at the end. If you do not know the answer,
just say that you do not know, do not try to make up an answer.
Context: {context}
Question: {question}
Helpful Answer:"""

# this is the default prompt template for if the user wants to summarize the context before qa prompt
DEFAULT_SUMMARIZATION_PROMPT_TEMPLATE = """
Summarize the following texts for me:
{context}

When summarizing, please keep the following in mind the following question:
{question}
"""

DEFAULT_CHUNK_SIZE = 750
DEFAULT_CHUNK_OVERLAP = 250
DEFAULT_VECTOR_STORE_NAME = "chromadb"
DEFAULT_VECTOR_STORE_COLLECTION_NAME = "collection"
MAX_EMBEDDINGS_BATCH_SIZE = 2000

chromadb = get_chromadb()


def is_valid_store(name) -> bool:
    return name in SUPPORTED_VECTOR_STORES


class VectorStoreFactory:
    """Factory class for vector stores"""

    @staticmethod
    def get_vectorstore_class(name) -> Union[FAISS, Chroma, VectorStore]:

        if not isinstance(name, str):
            raise TypeError("name must be a string")

        if not is_valid_store(name):
            raise ValueError(f"Invalid vector store {name}")

        if name == "faiss":
            return FAISS

        if name == "chromadb":
            return Chroma


def get_chroma_client(persist_directory: str) -> chromadb.PersistentClient:
    """Get Chroma client"""
    return chromadb.PersistentClient(path=persist_directory)


def get_available_writer_model_ids(args: ValidationInfo) -> list:
    """Get available writer LLM model ids"""

    args = args.data

    writer_client = writer.Writer(
        api_key=args["writer_api_key"],
        organization_id=args["writer_org_id"],
    )

    res = writer_client.models.list(organization_id=args["writer_org_id"])

    available_models_dict = json.loads(res.raw_response.text)

    return [model["id"] for model in available_models_dict["models"]]


def get_available_openai_model_ids(args: ValidationInfo) -> list:
    """Get available openai LLM model ids"""

    args = args.data

    models = openai.OpenAI(api_key=args["openai_api_key"], base_url=args.get("base_url")).models.list().data

    return [models.id for models in models]


@dataclass
class PersistedVectorStoreSaverConfig:
    vector_store_name: str
    persist_directory: str
    collection_name: str
    vector_store: VectorStore


@dataclass
class PersistedVectorStoreLoaderConfig:
    vector_store_name: str
    embeddings_model: Embeddings
    persist_directory: str
    collection_name: str


class PersistedVectorStoreSaver:
    """Saves vector store to disk"""

    def __init__(self, config: PersistedVectorStoreSaverConfig):
        self.config = config

    def save_vector_store(self, vector_store: VectorStore):
        method_name = f"save_{self.config.vector_store_name}"
        getattr(self, method_name)(vector_store)

    def save_chromadb(self, vector_store: Chroma):
        """Save Chroma vector store to disk"""
        # no need to save chroma vector store to disk, auto save
        pass

    def save_faiss(self, vector_store: FAISS):
        vector_store.save_local(
            folder_path=self.config.persist_directory,
            index_name=self.config.collection_name,
        )


class PersistedVectorStoreLoader:
    """Loads vector store from disk"""

    def __init__(self, config: PersistedVectorStoreLoaderConfig):
        self.config = config

    def load_vector_store_client(
            self,
            vector_store: str,
    ):
        """Load vector store from the persisted vector store"""

        if vector_store == "chromadb":

            return Chroma(
                collection_name=self.config.collection_name,
                embedding_function=self.config.embeddings_model,
                client=get_chroma_client(self.config.persist_directory),
            )

        elif vector_store == "faiss":

            return FAISS.load_local(
                folder_path=self.config.persist_directory,
                embeddings=self.config.embeddings_model,
                index_name=self.config.collection_name,
                allow_dangerous_deserialization=True
            )

        else:
            raise NotImplementedError(f"{vector_store} client is not yet supported")

    def load_vector_store(self) -> VectorStore:
        """Load vector store from the persisted vector store"""
        method_name = f"load_{self.config.vector_store_name}"
        return getattr(self, method_name)()

    def load_chromadb(self) -> Chroma:
        """Load Chroma vector store from the persisted vector store"""
        return self.load_vector_store_client(vector_store="chromadb")

    def load_faiss(self) -> FAISS:
        """Load FAISS vector store from the persisted vector store"""
        return self.load_vector_store_client(vector_store="faiss")


class LLMParameters(BaseModel):
    """Model parameters for the LLM API interface"""

    llm_name: str = Field(default_factory=str, title="LLM API name")
    max_tokens: int = Field(default=100, title="max tokens in response")
    temperature: float = Field(default=0.0, title="temperature")
    base_url: Optional[str] = None
    top_p: float = 1
    best_of: int = 5
    stop: Optional[List[str]] = None

    class Config:
        extra = Extra.forbid
        arbitrary_types_allowed = True
        use_enum_values = True
        protected_namespaces = ()


class OpenAIParameters(LLMParameters):
    """Model parameters for the LLM API interface"""

    openai_api_key: str
    model_id: str = Field(default="gpt-3.5-turbo-instruct", title="model name")
    n: int = Field(default=1, title="number of responses to return")

    @field_validator("model_id", mode="after")
    def openai_model_must_be_supported(cls, v, values):
        supported_models = get_available_openai_model_ids(values)
        if v not in supported_models:
            raise InvalidOpenAIModel(
                f"'model_id' must be one of {supported_models}, got {v}"
            )
        return v


class WriterLLMParameters(LLMParameters):
    """Model parameters for the Writer LLM API interface"""

    writer_api_key: str
    writer_org_id: Optional[str] = None
    model_id: str = "palmyra-x"
    callbacks: List[StreamingStdOutCallbackHandler] = [StreamingStdOutCallbackHandler()]
    verbose: bool = False

    @field_validator("model_id")
    def writer_model_must_be_supported(cls, v, values):
        supported_models = get_available_writer_model_ids(values)
        if v not in supported_models:
            raise InvalidWriterModel(
                f"'model_id' must be one of {supported_models}, got {v}"
            )
        return v


class LLMLoader(BaseModel):
    llm_config: dict
    config_dict: dict = None

    def load_llm(self) -> Union[Writer, partial]:
        """Load LLM"""
        method_name = f"load_{self.llm_config['llm_name']}_llm"
        self.config_dict = self.llm_config.copy()
        self.config_dict.pop("llm_name")
        return getattr(self, method_name)()

    def load_writer_llm(self) -> Writer:
        """Load Writer LLM API interface"""
        return Writer(**self.config_dict)

    def load_openai_llm(self) -> partial:
        """Load OpenAI LLM API interface"""
        client = openai.OpenAI(api_key=self.config_dict["openai_api_key"], base_url=self.config_dict["base_url"])
        config = self.config_dict.copy()
        keys_to_remove = ["openai_api_key", "base_url"]
        for key in keys_to_remove:
            config.pop(key)
        config["model"] = config.pop("model_id")

        return partial(client.completions.create, **config)


class RAGBaseParameters(BaseModel):
    """Base model parameters for RAG Handler"""

    llm_params: Any
    vector_store_folder_name: str
    input_column: str
    use_gpu: bool = False
    embeddings_batch_size: int = MAX_EMBEDDINGS_BATCH_SIZE  # not used, leaving in place to prevent breaking changes
    prompt_template: str = DEFAULT_QA_PROMPT_TEMPLATE
    chunk_size: int = DEFAULT_CHUNK_SIZE
    chunk_overlap: int = DEFAULT_CHUNK_OVERLAP
    url: Optional[Union[str, List[str]]] = None
    url_column_name: Optional[str] = None
    run_embeddings: Optional[bool] = True
    top_k: int = 4
    embeddings_model: Optional[Embeddings] = None
    embeddings_model_name: str = DEFAULT_EMBEDDINGS_MODEL
    context_columns: Optional[Union[List[str], str]] = None
    vector_store_name: str = DEFAULT_VECTOR_STORE_NAME
    vector_store: Optional[VectorStore] = None
    collection_name: str = DEFAULT_VECTOR_STORE_COLLECTION_NAME
    summarize_context: bool = True
    summarization_prompt_template: str = DEFAULT_SUMMARIZATION_PROMPT_TEMPLATE
    vector_store_storage_path: Optional[str] = Field(
        default=None, title="don't use this field, it's for internal use only"
    )

    class Config:
        extra = Extra.forbid
        arbitrary_types_allowed = True
        use_enum_values = True

    @field_validator("prompt_template")
    def prompt_format_must_be_valid(cls, v):
        if "{context}" not in v or "{question}" not in v:
            raise InvalidPromptTemplate(
                "prompt_template must contain {context} and {question}"
                f"\n For example, {DEFAULT_QA_PROMPT_TEMPLATE}"
            )
        return v

    @field_validator("vector_store_name")
    def name_must_be_lower(cls, v):
        return v.lower()

    @field_validator("vector_store_name")
    def vector_store_must_be_supported(cls, v):
        if not is_valid_store(v):
            raise UnsupportedVectorStore(
                f"we don't support {v}. currently we only support {', '.join(str(v) for v in SUPPORTED_VECTOR_STORES)} vector store"
            )
        return v


class RAGHandlerParameters(RAGBaseParameters):
    """Model parameters for create model"""

    llm_type: str
    llm_params: LLMParameters

    @field_validator("llm_type")
    def llm_type_must_be_supported(cls, v):
        if v not in SUPPORTED_LLMS:
            raise UnsupportedLLM(f"'llm_type' must be one of {SUPPORTED_LLMS}, got {v}")
        return v


class DfLoader(DataFrameLoader):
    """
    override the load method of langchain.document_loaders.DataFrameLoaders to ignore rows with 'None' values
    """

    def __init__(self, data_frame: pd.DataFrame, page_content_column: str):
        super().__init__(data_frame=data_frame, page_content_column=page_content_column)
        self._data_frame = data_frame
        self._page_content_column = page_content_column

    def load(self) -> List[Document]:
        """Loads the dataframe as a list of documents"""
        documents = []
        for n_row, frame in self._data_frame[self._page_content_column].items():
            if pd.notnull(frame):
                # ignore rows with None values
                column_name = self._page_content_column

                document_contents = frame

                documents.append(
                    Document(
                        page_content=document_contents,
                        metadata={
                            "source": "dataframe",
                            "row": n_row,
                            "column": column_name,
                        },
                    )
                )
        return documents


def df_to_documents(
        df: pd.DataFrame,
        page_content_columns: Union[List[str], str],
        url_column_name: str = None,
) -> List[Document]:
    """Converts a given dataframe to a list of documents"""
    documents = []

    if isinstance(page_content_columns, str):
        page_content_columns = [page_content_columns]

    for _, page_content_column in enumerate(page_content_columns):
        if page_content_column not in df.columns.tolist():
            raise ValueError(
                f"page_content_column {page_content_column} not in dataframe columns"
            )
        if url_column_name is not None and page_content_column == url_column_name:
            documents.extend(url_to_documents(df[page_content_column].tolist()))
            continue

        loader = DfLoader(data_frame=df, page_content_column=page_content_column)
        documents.extend(loader.load())

    return documents


def url_to_documents(urls: Union[List[str], str]) -> List[Document]:
    """Converts a given url to a document"""
    documents = []
    if isinstance(urls, str):
        urls = [urls]

    for url in urls:
        response = requests.get(url, headers=None).text
        html_to_text = html2text.html2text(response)
        documents.append(Document(page_content=html_to_text, metadata={"source": url}))

    return documents


@lru_cache()
def load_embeddings_model(embeddings_model_name, use_gpu=False):
    """Load embeddings model from Hugging Face Hub"""
    try:
        model_kwargs = dict(device="cuda" if use_gpu else "cpu")
        embedding_model = HuggingFaceEmbeddings(
            model_name=embeddings_model_name, model_kwargs=model_kwargs
        )
    except ValueError:
        raise ValueError(
            f"The {embeddings_model_name}  is not supported, please select a valid option from Hugging Face Hub!"
        )
    return embedding_model


def on_create_build_llm_params(
        args: dict, llm_config_class: Union[WriterLLMParameters, OpenAIParameters]
) -> Dict:
    """build llm params from create args"""

    llm_params = {"llm_name": args["llm_type"]}

    for param in llm_config_class.model_fields.keys():
        if param in args:
            llm_params[param] = args.pop(param)

    return llm_params


def build_llm_params(args: dict, update=False) -> Dict:
    """build llm params from args"""

    if args["llm_type"] == "writer":
        llm_config_class = WriterLLMParameters
    elif args["llm_type"] == "openai":
        llm_config_class = OpenAIParameters
    else:
        raise UnsupportedLLM(
            f"'llm_type' must be one of {SUPPORTED_LLMS}, got {args['llm_type']}"
        )

    if not args.get("llm_params"):
        # for create method only
        llm_params = on_create_build_llm_params(args, llm_config_class)
    else:
        # for predict method only
        llm_params = args.pop("llm_params")
    if update:
        # for update method only
        args["llm_params"] = llm_params
        return args

    args["llm_params"] = llm_config_class(**llm_params)

    return args



import json
import re
from pydantic import BaseModel, Field
from typing import Any, List, Optional

from langchain.chains.llm import LLMChain
from langchain_core.callbacks.manager import CallbackManagerForRetrieverRun
from langchain_core.documents.base import Document
from langchain_core.embeddings import Embeddings
from langchain_core.exceptions import OutputParserException
from langchain_core.language_models.chat_models import BaseChatModel
from langchain_core.output_parsers import PydanticOutputParser
from langchain_core.prompts import PromptTemplate
from langchain_core.retrievers import BaseRetriever

from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE
from mindsdb.integrations.libs.response import HandlerResponse
from mindsdb.integrations.libs.vectordatabase_handler import DistanceFunction, VectorStoreHandler
from mindsdb.integrations.utilities.rag.settings import LLMExample, MetadataSchema, SearchKwargs
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class MetadataFilter(BaseModel):
    '''Represents an LLM generated metadata filter to apply to a PostgreSQL query.'''
    attribute: str = Field(description="Database column to apply filter to")
    comparator: str = Field(description="PostgreSQL comparator to use to filter database column")
    value: Any = Field(description="Value to use to filter database column")


class MetadataFilters(BaseModel):
    '''List of LLM generated metadata filters to apply to a PostgreSQL query.'''
    filters: List[MetadataFilter] = Field(description="List of PostgreSQL metadata filters to apply for user query")


class SQLRetriever(BaseRetriever):
    '''Retriever that uses a LLM to generate pgvector queries to do similarity search with metadata filters.

    How it works:

    1. Use a LLM to rewrite the user input to something more suitable for retrieval. For example:
    "Show me documents containing how to finetune a LLM please" --> "how to finetune a LLM"

    2. Use a LLM to generate structured metadata filters based on the user input. Provided
    metadata schemas & examples are used as additional context.

    3. Generate a prepared PostgreSQL query from the structured metadata filters.

    4. Actually execute the query against our vector database to retrieve documents & return them.
    '''
    fallback_retriever: BaseRetriever
    vector_store_handler: VectorStoreHandler
    metadata_schemas: Optional[List[MetadataSchema]] = None
    examples: Optional[List[LLMExample]] = None

    rewrite_prompt_template: str
    metadata_filters_prompt_template: str
    embeddings_model: Embeddings
    num_retries: int
    embeddings_table: str
    source_table: str
    source_id_column: str = 'Id'
    distance_function: DistanceFunction
    search_kwargs: SearchKwargs

    llm: BaseChatModel

    def _prepare_metadata_prompt(self) -> PromptTemplate:
        base_prompt_template = PromptTemplate(
            input_variables=['format_instructions', 'schema', 'examples', 'input', 'embeddings'],
            template=self.metadata_filters_prompt_template
        )
        schema_prompt_str = ''
        if self.metadata_schemas is not None:
            for i, schema in enumerate(self.metadata_schemas):
                column_mapping = {}
                for column in schema.columns:
                    column_mapping[column.name] = {
                        'type': column.type,
                        'description': column.description,
                    }
                    if column.values is not None:
                        column_mapping[column.name]['values'] = column.values
                column_mapping_json_str = json.dumps(column_mapping, indent=4)
                schema_str = f'''{i+1}. {schema.table} - {schema.description}

Columns:
```json
{column_mapping_json_str}
```

'''
                schema_prompt_str += schema_str

        examples_prompt_str = ''
        if self.examples is not None:
            for i, example in enumerate(self.examples):
                example_str = f'''{i + 1}. User input: "{example.input}"

Output:
{example.output}

'''
                examples_prompt_str += example_str
        return base_prompt_template.partial(
            schema=schema_prompt_str,
            examples=examples_prompt_str
        )

    def _prepare_retrieval_query(self, query: str) -> str:
        rewrite_prompt = PromptTemplate(
            input_variables=['input'],
            template=self.rewrite_prompt_template
        )
        rewrite_chain = LLMChain(llm=self.llm, prompt=rewrite_prompt)
        return rewrite_chain.predict(input=query)

    def _prepare_pgvector_query(self, metadata_filters: List[MetadataFilter]) -> str:
        # Base select JOINed with document source table.
        base_query = f'''SELECT * FROM {self.embeddings_table} AS e INNER JOIN {self.source_table} AS s ON (e.metadata->>'original_row_id')::int = s."{self.source_id_column}" '''
        col_to_schema = {}
        if not self.metadata_schemas:
            return ''
        for schema in self.metadata_schemas:
            for col in schema.columns:
                col_to_schema[col.name] = schema
        joined_schemas = set()
        for filter in metadata_filters:
            # Join schemas before filtering.
            schema = col_to_schema.get(filter.attribute)
            if schema is None or schema.table in joined_schemas or schema.table == self.source_table:
                continue
            joined_schemas.add(schema.table)
            base_query += schema.join + ' '
        # Actually construct WHERE conditions from metadata filters.
        if metadata_filters:
            base_query += 'WHERE '
        for i, filter in enumerate(metadata_filters):
            value = filter.value
            if isinstance(value, str):
                value = f"'{value}'"
            base_query += f'"{filter.attribute}" {filter.comparator} {value}'
            if i < len(metadata_filters) - 1:
                base_query += ' AND '
        base_query += f" ORDER BY e.embeddings {self.distance_function.value[0]} '{{embeddings}}' LIMIT {self.search_kwargs.k};"
        return base_query

    def _generate_metadata_filters(self, query: str) -> List[MetadataFilter]:
        parser = PydanticOutputParser(pydantic_object=MetadataFilters)
        metadata_prompt = self._prepare_metadata_prompt()
        metadata_filters_chain = LLMChain(llm=self.llm, prompt=metadata_prompt)
        metadata_filters_output = metadata_filters_chain.predict(
            format_instructions=parser.get_format_instructions(),
            input=query
        )
        # If the LLM outputs raw JSON, use it as-is.
        # If the LLM outputs anything including a json markdown section, use the last one.
        json_markdown_output = re.findall(r'```json.*```', metadata_filters_output, re.DOTALL)
        if json_markdown_output:
            metadata_filters_output = json_markdown_output[-1]
            # Clean the json tags.
            metadata_filters_output = metadata_filters_output[7:]
            metadata_filters_output = metadata_filters_output[:-3]
        metadata_filters = parser.invoke(metadata_filters_output)
        return metadata_filters.filters

    def _prepare_and_execute_query(self, query: str, embeddings_str: str) -> HandlerResponse:
        try:
            metadata_filters = self._generate_metadata_filters(query)
            checked_sql_query = self._prepare_pgvector_query(metadata_filters)
            checked_sql_query_with_embeddings = checked_sql_query.format(embeddings=embeddings_str)
            return self.vector_store_handler.native_query(checked_sql_query_with_embeddings)
        except OutputParserException as e:
            logger.warning(f'LLM failed to generate structured metadata filters: {str(e)}')
            return HandlerResponse(RESPONSE_TYPE.ERROR, error_message=str(e))
        except Exception as e:
            logger.warning(f'Failed to prepare and execute SQL query from structured metadata: {str(e)}')
            return HandlerResponse(RESPONSE_TYPE.ERROR, error_message=str(e))

    def _get_relevant_documents(
        self, query: str, *, run_manager: CallbackManagerForRetrieverRun
    ) -> List[Document]:
        # Rewrite query to be suitable for retrieval.
        retrieval_query = self._prepare_retrieval_query(query)
        # Embed the rewritten retrieval query & include it in the similarity search pgvector query.
        embedded_query = self.embeddings_model.embed_query(retrieval_query)
        # Actually execute the similarity search with metadata filters.
        document_response = self._prepare_and_execute_query(retrieval_query, str(embedded_query))
        num_retries = 0
        while num_retries < self.num_retries:
            if document_response.resp_type != RESPONSE_TYPE.ERROR and len(document_response.data_frame) > 0:
                # Successfully retrieved documents.
                break
            if document_response.resp_type == RESPONSE_TYPE.ERROR:
                # LLMs won't always generate structured metadata so we should have a fallback after retrying.
                logger.info(f'SQL Retriever query failed with error {document_response.error_message}')
            elif len(document_response.data_frame) == 0:
                logger.info('No documents retrieved from SQL Retriever query')

            document_response = self._prepare_and_execute_query(retrieval_query, str(embedded_query))
            num_retries += 1
            if num_retries >= self.num_retries:
                logger.info('Using fallback retriever in SQL retriever.')
                return self.fallback_retriever._get_relevant_documents(retrieval_query, run_manager=run_manager)

        document_df = document_response.data_frame
        retrieved_documents = []
        for _, document_row in document_df.iterrows():
            retrieved_documents.append(Document(
                document_row.get('content', ''),
                metadata=document_row.get('metadata', {})
            ))
        if retrieved_documents:
            return retrieved_documents
        # If the SQL query constructed did not return any documents, fallback.
        logger.info('No documents returned from SQL retriever. using fallback retriever.')
        return self.fallback_retriever._get_relevant_documents(retrieval_query, run_manager=run_manager)



from typing import List

import pandas as pd
from langchain_core.documents import Document
from langchain_core.embeddings import Embeddings


def df_to_documents(df: pd.DataFrame, content_column_name: str) -> List[Document]:
    """
    Given a dataframe, convert it to a list of documents.

    :param df: pd.DataFrame
    :param content_column_name: str

    :return: List[Document]
    """
    documents = []
    for _, row in df.iterrows():
        metadata = row.to_dict()
        page_content = metadata.pop(content_column_name)
        documents.append(Document(page_content=page_content, metadata=metadata))
    return documents


def documents_to_df(content_column_name: str,
                    documents: List[Document],
                    embedding_model: Embeddings = None,
                    with_embeddings: bool = False) -> pd.DataFrame:
    """
    Given a list of documents, convert it to a dataframe.

    :param content_column_name: str
    :param documents: List[Document]
    :param embedding_model: Embeddings
    :param with_embeddings: bool

    :return: pd.DataFrame
    """
    df = pd.DataFrame([doc.metadata for doc in documents])

    df[content_column_name] = [doc.page_content for doc in documents]

    if 'date' in df.columns:
        df['date'] = pd.to_datetime(df['date'], errors='coerce')

    # Reordering the columns to have the content column first.
    df = df[[content_column_name] + [col for col in df.columns if col != content_column_name]]

    if with_embeddings:
        df["embeddings"] = embedding_model.embed_documents(df[content_column_name].tolist())

    return df



from mindsdb.api.mongo.classes import Responder


class Responce(Responder):
    def when(self, query):
        return (
            'buildinfo' in query or 'buildInfo' in query
        )

    result = {
        'version': '3.6.8',
        'versionArray': [3, 6, 8, 0],
        'ok': 1
    }


responder = Responce()



import datetime as dt
import queue
import random
import threading
import time

from mindsdb.interfaces.jobs.jobs_controller import JobsExecutor
from mindsdb.interfaces.storage import db
from mindsdb.utilities import log
from mindsdb.utilities.config import Config
from mindsdb.utilities.sentry import sentry_sdk  # noqa: F401

logger = log.getLogger(__name__)


def execute_async(q_in, q_out):

    while True:
        task = q_in.get()

        if task["type"] != "task":
            return

        record_id = task["record_id"]
        history_id = task["history_id"]

        executor = JobsExecutor()
        try:
            executor.execute_task_local(record_id, history_id)
        except (KeyboardInterrupt, SystemExit):
            q_out.put(True)
            raise

        except Exception:
            db.session.rollback()

        q_out.put(True)


class Scheduler:
    def __init__(self, config=None):
        self.config = config

        self.q_in = queue.Queue()
        self.q_out = queue.Queue()
        self.work_thread = threading.Thread(
            target=execute_async, args=(self.q_in, self.q_out)
        )
        self.work_thread.start()

    def __del__(self):
        self.stop_thread()

    def stop_thread(self):
        self.q_in.put({"type": "exit"})

    def scheduler_monitor(self):
        check_interval = self.config.get("jobs", {}).get("check_interval", 30)

        while True:

            logger.debug("Scheduler check timetable")
            try:
                self.check_timetable()
            except (SystemExit, KeyboardInterrupt):
                raise
            except Exception as e:
                logger.error(e)

            # different instances should start in not the same time

            time.sleep(check_interval + random.randint(1, 10))

    def check_timetable(self):
        executor = JobsExecutor()

        exec_method = self.config.get("jobs", {}).get("executor", "local")

        for record in executor.get_next_tasks():
            logger.info(f"Job execute: {record.name}({record.id})")
            self.execute_task(record.id, exec_method)

        db.session.remove()

    def execute_task(self, record_id, exec_method):

        executor = JobsExecutor()
        if exec_method == "local":
            history_id = executor.lock_record(record_id)
            if history_id is None:
                # db.session.remove()
                logger.info(f"Unable create history record for {record_id}, is locked?")
                return

            # run in thread

            self.q_in.put(
                {
                    "type": "task",
                    "record_id": record_id,
                    "history_id": history_id,
                }
            )

            while True:
                try:
                    self.q_out.get(timeout=3)
                    break
                except queue.Empty:
                    # update last date:
                    history_record = db.JobsHistory.query.get(history_id)
                    history_record.updated_at = dt.datetime.now()
                    db.session.commit()

        else:
            # TODO add microservice mode
            raise NotImplementedError()

    def start(self):

        config = Config()
        db.init()
        self.config = config

        logger.info("Scheduler starts")

        try:
            self.scheduler_monitor()
        except (KeyboardInterrupt, SystemExit):

            self.stop_thread()
            pass


def start(verbose=False):
    logger.info("Jobs API is starting..")
    scheduler = Scheduler()

    scheduler.start()


if __name__ == "__main__":
    start()



import pandas as pd
import requests
from mindsdb.integrations.libs.api_handler import APITable
from mindsdb_sql_parser import ast
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions


class BaseResultsTable(APITable):
    def select(self, query: ast.Select) -> pd.DataFrame:
        """
        Selects data from the results table and returns it as a pandas DataFrame.

        Args:
            query (ast.Select): The SQL query to be executed.

        Returns:
            pandas.DataFrame: A pandas DataFrame containing the selected data.
        """
        base_url = self.handler.connect()

        params = {'access_key': self.handler.access_key}
        conditions = extract_comparison_conditions(query.where)
        params.update({condition[1]: condition[2] for condition in conditions if condition[0] == '='})

        if 'query' not in params:
            raise ValueError('Query is missing in the SQL query')
        if 'type' not in params and hasattr(self, 'default_type'):
            params['type'] = self.default_type
        try:
            api_response = requests.get(base_url, params=params)
            api_response.raise_for_status()  # raises HTTPError for bad responses
            api_result = api_response.json()
        except requests.exceptions.HTTPError as e:
            raise SystemError(f"HTTP error occurred: {e.response.status_code} - {e.response.reason}")
        except requests.exceptions.ConnectionError as e:
            raise SystemError(f"Connection error occurred: {str(e)}")
        except requests.exceptions.Timeout as e:
            raise SystemError(f"Request timeout: {str(e)}")
        except requests.exceptions.RequestException as e:
            raise SystemError(f"Request exception occurred: {str(e)}")
        except ValueError as e:
            raise SystemError(f"Failed to parse JSON response: {str(e)}")

        results = api_result.get(self.results_key, [])
        processed_results = [self.extract_data(result) for result in results]

        if len(processed_results) == 0:
            columns = self.get_columns()
            empty_data = {col: ["No results found"] for col in columns}
            return pd.DataFrame(empty_data, columns=columns)

        result_df = pd.DataFrame(processed_results)
        result_df = self.filter_columns(result_df, query)
        return result_df

    def extract_data(self, data):
        """
        Extracts the required data from the result.

        Args:
            data (dict): The result data.

        Returns:
            dict: A dictionary containing the extracted data.
        """
        raise NotImplementedError("Subclasses must implement this method.")

    def filter_columns(self, result: pd.DataFrame, query: ast.Select = None):
        """
        Filters the columns of the result DataFrame.

        Args:
            result (pandas.DataFrame): The result DataFrame.
            query (ast.Select): The SQL query to be executed.

        Returns:
            pandas.DataFrame: A pandas DataFrame containing the filtered data.
        """
        columns = []
        if query is not None:
            for target in query.targets:
                if isinstance(target, ast.Star):
                    columns = self.get_columns()
                    break
                elif isinstance(target, ast.Identifier):
                    columns.append(target.parts[-1])
                else:
                    raise NotImplementedError
        else:
            columns = self.get_columns()

        columns = [name.lower() for name in columns]

        if len(result) == 0:
            result = pd.DataFrame([], columns=columns)
        else:
            for col in set(columns) & set(result.columns) ^ set(columns):
                result[col] = None

            result = result[columns]

        if query is not None and query.limit is not None:
            result = result.head(query.limit.value)

        return result


class OrganicResultsTable(BaseResultsTable):
    results_key = 'organic_results'

    def extract_data(self, organic):
        return {
            'position': organic.get('position'),
            'title': organic.get('title'),
            'url': organic.get('url'),
            'domain': organic.get('domain'),
            'displayed_url': organic.get('displayed_url'),
            'snippet': organic.get('snippet'),
            'cached_page_url': organic.get('cached_page_url'),
            'related_pages_url': organic.get('related_pages_url'),
            'prerender': organic.get('prerender'),
            'sitelinks': self._extract_sitelinks(organic.get('sitelinks')),
            'rich_snippet': self._extract_rich_snippet(organic.get('rich_snippet'))
        }

    def _extract_sitelinks(self, sitelinks):
        if not sitelinks:
            return None
        return {
            'inline': [{'title': link['title'], 'url': link['url']} for link in sitelinks.get('inline', [])],
            'expanded': [{'title': link['title'], 'url': link['url']} for link in sitelinks.get('expanded', [])]
        }

    def _extract_rich_snippet(self, rich_snippet):
        if not rich_snippet:
            return None
        snippet_type = 'top' if 'top' in rich_snippet else 'bottom'
        return {
            'detected_extensions': rich_snippet.get(snippet_type, {}).get('detected_extensions', []),
            'extensions': rich_snippet.get(snippet_type, {}).get('extensions', [])
        }

    def get_columns(self):
        return [
            'position',
            'title',
            'url',
            'domain',
            'displayed_url',
            'snippet',
            'cached_page_url',
            'related_pages_url',
            'prerender',
            'sitelinks',
            'rich_snippet'
        ]


class ImageResultsTable(BaseResultsTable):
    results_key = 'image_results'
    default_type = 'images'

    def extract_data(self, image):
        return {
            'position': image.get('position'),
            'title': image.get('title'),
            'width': image.get('width'),
            'height': image.get('height'),
            'image_url': image.get('image_url'),
            'type': image.get('type'),
            'url': image.get('url'),
            'source': image.get('source')
        }

    def get_columns(self):
        return [
            'position',
            'title',
            'width',
            'height',
            'image_url',
            'type',
            'url',
            'source'
        ]


class VideoResultsTable(BaseResultsTable):
    results_key = 'video_results'
    default_type = 'videos'

    def extract_data(self, video):
        return {
            'position': video.get('position'),
            'title': video.get('title'),
            'url': video.get('url'),
            'displayed_url': video.get('displayed_url'),
            'uploaded': video.get('uploaded'),
            'snippet': video.get('snippet'),
            'length': video.get('length')
        }

    def get_columns(self):
        return [
            'position',
            'title',
            'url',
            'displayed_url',
            'uploaded',
            'snippet',
            'length'
        ]


class NewsResultsTable(BaseResultsTable):
    results_key = 'news_results'
    default_type = 'news'

    def extract_data(self, news):
        return {
            'position': news.get('position'),
            'title': news.get('title'),
            'url': news.get('url'),
            'source_name': news.get('source_name'),
            'uploaded': news.get('uploaded'),
            'uploaded_utc': news.get('uploaded_utc'),
            'snippet': news.get('snippet'),
            'thumbnail_url': news.get('thumbnail_url')
        }

    def get_columns(self):
        return [
            'position',
            'title',
            'url',
            'source_name',
            'uploaded',
            'uploaded_utc',
            'snippet',
            'thumbnail_url'
        ]


class ShoppingResultsTable(BaseResultsTable):
    results_key = 'shopping_results'
    default_type = 'shopping'

    def extract_data(self, shopping):
        return {
            'position': shopping.get('position'),
            'title': shopping.get('title'),
            'url': shopping.get('url')
        }

    def get_columns(self):
        return [
            'position',
            'title',
            'url'
        ]



import requests
import time
from typing import Dict, Generator, List, Optional, Text, Union

from mindsdb.utilities import log

logger = log.getLogger(__name__)


class MSGraphAPIBaseClient:
    """
    The base class for the Microsoft Graph API clients.
    This class contains common methods for accessing the Microsoft Graph API.

    Attributes:
        MICROSOFT_GRAPH_BASE_API_URL (Text): The base URL of the Microsoft Graph API.
        MICROSOFT_GRAPH_API_VERSION (Text): The version of the Microsoft Graph API.
        PAGINATION_COUNT (Optional[int]): The number of items to retrieve per request.
    """
    MICROSOFT_GRAPH_BASE_API_URL: Text = "https://graph.microsoft.com/"
    MICROSOFT_GRAPH_API_VERSION: Text = "v1.0"
    PAGINATION_COUNT: Optional[int] = 20

    def __init__(self, access_token: Text) -> None:
        """
        Initializes the Microsoft Graph API client.

        Args:
            access_token (Text): The access token for authenticating the requests to the Microsoft Graph API.
        """
        self.access_token = access_token
        self._group_ids = None

    def _get_api_url(self, endpoint: Text) -> Text:
        """
        Constructs the API URL for the specified endpoint.

        Args:
            endpoint (Text): The endpoint of the Microsoft Graph API.

        Returns:
            Text: The fully constructed API URL.
        """
        api_url = f"{self.MICROSOFT_GRAPH_BASE_API_URL}{self.MICROSOFT_GRAPH_API_VERSION}/{endpoint}/"
        return api_url

    def _make_request(
        self,
        api_url: Text,
        params: Optional[Dict] = None,
        data: Optional[Dict] = None,
        method: Text = "GET"
    ) -> Union[Dict, object]:
        """
        Makes a request to the Microsoft Graph API.

        Args:
            api_url (Text): The API URL to make the request to.
            params (Optional[Dict]): The parameters to include in the request.
            data (Optional[Dict]): The data to include in the request.
            method (Text): The HTTP method to use for the request.

        Returns:
            Union[Dict, object]: The response content of the request.
        """
        headers = {"Authorization": f"Bearer {self.access_token}"}

        # Make the request to the Microsoft Graph API based on the method.
        if method == "GET":
            response = requests.get(api_url, headers=headers, params=params)
        elif method == "POST":
            response = requests.post(api_url, headers=headers, json=data)
        else:
            raise NotImplementedError(f"Method {method} not implemented")

        # Process the response.
        # If the response is a 429 (rate limit exceeded), wait for the specified time and retry.
        if response.status_code == 429:
            if "Retry-After" in response.headers:
                pause_time = float(response.headers["Retry-After"])
                time.sleep(pause_time)
                response = requests.get(api_url, headers=headers, params=params)

        # If the response is not successful, raise an exception.
        if response.status_code not in [200, 201]:
            raise requests.exceptions.RequestException(response.text)

        return response

    def fetch_paginated_data(self, endpoint: Text, params: Optional[Dict] = {}) -> Generator:
        """
        Fetches data from the Microsoft Graph API by making the specified request and handling pagination.

        Args:
            endpoint (str): The endpoint of the Microsoft Graph API to fetch data from.
            params (Optional[Dict]): The parameters to include in the request.

        Yields:
            List: The data fetched from the Microsoft Graph API.
        """
        api_url = self._get_api_url(endpoint)

        # Add the pagination count to the request parameters.
        if "$top" not in params:
            params["$top"] = self.PAGINATION_COUNT

        while api_url:
            # Make the initial request to the Microsoft Graph API.
            response = self._make_request(api_url, params)
            response_json = response.json()
            value = response.json()["value"]

            # Get the next page of data if pagination is enabled.
            params = None
            api_url = response_json.get("@odata.nextLink", "")
            yield value

    def fetch_data(self, endpoint: str, params: Optional[Dict] = {}) -> Union[List, Dict, bytes]:
        """
        Fetches data from the Microsoft Graph API by making the specified request.

        Args:
            endpoint (str): The endpoint of the Microsoft Graph API to fetch data from.
            params (Optional[Dict]): The parameters to include in the request.

        Returns:
            Union[List, Dict, bytes]: The data fetched from the Microsoft Graph API.
        """
        api_url = self._get_api_url(endpoint)

        response = self._make_request(api_url, params)

        # If the response content is a binary file or a TSV file, return the raw content.
        if response.headers["Content-Type"] in ("application/octet-stream", "text/plain",
                                                "text/tab-separated-values", "application/pdf"):
            return response.content
        # Otherwise, return the JSON content.
        else:
            return response.json()



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    persist_directory={
        "type": ARG_TYPE.STR,
        "description": "The uri of the database.",
        "required": True,
    },
    api_key={
        "type": ARG_TYPE.STR,
        "description": "If presented, connect to LanceDB cloud. Otherwise, connect to a database on file system or cloud storage.",
        "required": False,
        "secret": True
    },
    region={
        "type": ARG_TYPE.STR,
        "description": "The region to use for LanceDB Cloud.",
        "required": False,
    },
    host_override={
        "type": ARG_TYPE.STR,
        "description": "The override url for LanceDB Cloud.",
        "required": False,
    },
)

connection_args_example = OrderedDict(
    persist_directory="~/lancedb",
    api_key=None,
    region="us-west-2",
    host_override=None,
)



import enum
import inspect
from dataclasses import dataclass
from collections import defaultdict
from typing import List, Dict, Optional

from langchain_core.embeddings import Embeddings
from langchain_core.language_models import BaseChatModel
from mindsdb_sql_parser.ast import Select, BinaryOperation, Identifier, Constant, Star

from mindsdb.utilities import log
from mindsdb.utilities.cache import get_cache
from mindsdb.interfaces.storage import db
from mindsdb.interfaces.skills.sql_agent import SQLAgent
from mindsdb.integrations.libs.vectordatabase_handler import TableField


_DEFAULT_TOP_K_SIMILARITY_SEARCH = 5
_MAX_CACHE_SIZE = 1000

logger = log.getLogger(__name__)


class SkillType(enum.Enum):
    TEXT2SQL_LEGACY = 'text2sql'
    TEXT2SQL = 'sql'
    KNOWLEDGE_BASE = 'knowledge_base'
    RETRIEVAL = 'retrieval'


@dataclass
class SkillData:
    """Storage for skill's data

    Attributes:
        name (str): name of the skill
        type (str): skill's type (SkillType)
        params (dict): skill's attributes
        project_id (int): id of the project
        agent_tables_list (Optional[List[str]]): the restriction on available tables for an agent using the skill
    """
    name: str
    type: str
    params: dict
    project_id: int
    agent_tables_list: Optional[List[str]]

    @property
    def restriction_on_tables(self) -> Optional[Dict[str, set]]:
        """Schemas and tables which agent+skill may use. The result is intersections of skill's and agent's tables lists.

        Returns:
            Optional[Dict[str, set]]: allowed schemas and tables. Schemas - are keys in dict, tables - are values.
                if result is None, then there are no restrictions

        Raises:
            ValueError: if there is no intersection between skill's and agent's list.
                This means that all tables restricted for use.
        """
        def list_to_map(input: List) -> Dict:
            agent_tables_map = defaultdict(set)
            for x in input:
                if isinstance(x, str):
                    table_name = x
                    schema_name = None
                elif isinstance(x, dict):
                    table_name = x['table']
                    schema_name = x.get('schema')
                else:
                    raise ValueError(f'Unexpected value in tables list: {x}')
                agent_tables_map[schema_name].add(table_name)
            return agent_tables_map

        agent_tables_map = list_to_map(self.agent_tables_list or [])
        skill_tables_map = list_to_map(self.params.get('tables', []))

        if len(agent_tables_map) > 0 and len(skill_tables_map) > 0:
            if len(set(agent_tables_map) & set(skill_tables_map)) == 0:
                raise ValueError("Skill's and agent's allowed tables list have no shared schemas.")

            intersection_tables_map = defaultdict(set)
            has_intersection = False
            for schema_name in agent_tables_map:
                if schema_name not in skill_tables_map:
                    continue
                intersection_tables_map[schema_name] = agent_tables_map[schema_name] & skill_tables_map[schema_name]
                if len(intersection_tables_map[schema_name]) > 0:
                    has_intersection = True
            if has_intersection is False:
                raise ValueError("Skill's and agent's allowed tables list have no shared tables.")
            return intersection_tables_map
        if len(skill_tables_map) > 0:
            return skill_tables_map
        if len(agent_tables_map) > 0:
            return agent_tables_map
        return None


class SkillToolController:
    def __init__(self):
        self.command_executor = None

    def get_command_executor(self):
        if self.command_executor is None:
            from mindsdb.api.executor.command_executor import ExecuteCommands
            from mindsdb.api.executor.controllers import SessionController  # Top-level import produces circular import in some cases TODO: figure out a fix without losing runtime improvements (context: see #9304)  # noqa

            sql_session = SessionController()
            sql_session.database = 'mindsdb'

            self.command_executor = ExecuteCommands(sql_session)
        return self.command_executor

    def _make_text_to_sql_tools(self, skills: List[db.Skills], llm) -> List:
        '''
           Uses SQLAgent to execute tool
        '''
        # To prevent dependency on Langchain unless an actual tool uses it.
        try:
            from mindsdb.interfaces.agents.mindsdb_database_agent import MindsDBSQL
            from mindsdb.interfaces.skills.custom.text2sql.mindsdb_sql_toolkit import MindsDBSQLToolkit
            from langchain_community.tools.sql_database.tool import QuerySQLDataBaseTool
        except ImportError:
            raise ImportError(
                'To use the text-to-SQL skill, please install langchain with `pip install mindsdb[langchain]`')

        command_executor = self.get_command_executor()

        tables_list = []
        for skill in skills:
            database = skill.params['database']
            restriction_on_tables = skill.restriction_on_tables
            if restriction_on_tables is None:
                handler = command_executor.session.integration_controller.get_data_handler(database)
                if 'all' in inspect.signature(handler.get_tables).parameters:
                    response = handler.get_tables(all=True)
                else:
                    response = handler.get_tables()
                # no restrictions
                if 'table_schema' in response.data_frame.columns:
                    for _, row in response.data_frame.iterrows():
                        tables_list.append(f"{database}.{row['table_schema']}.{row['table_name']}")
                else:
                    for _, row in response.data_frame.iterrows():
                        tables_list.append(f"{database}.{row['table_name']}")
                continue
            for schema_name, tables in restriction_on_tables.items():
                for table in tables:
                    if schema_name is None:
                        tables_list.append(f'{database}.{table}')
                    else:
                        tables_list.append(f'{database}.{schema_name}.{table}')

        sql_agent = SQLAgent(
            command_executor=command_executor,
            databases=list(set(s.params['database'] for s in skills)),
            databases_struct={
                skill.params['database']: skill.restriction_on_tables
                for skill in skills
            },
            include_tables=tables_list,
            ignore_tables=None,
            sample_rows_in_table_info=3,
            cache=get_cache('agent', max_size=_MAX_CACHE_SIZE)
        )
        db = MindsDBSQL.custom_init(
            sql_agent=sql_agent
        )

        # Users probably don't need to configure this for now.
        sql_database_tools = MindsDBSQLToolkit(db=db, llm=llm).get_tools()
        descriptions = []
        for skill in skills:
            description = skill.params.get('description', '')
            if description:
                descriptions.append(description)

        for i, tool in enumerate(sql_database_tools):
            if isinstance(tool, QuerySQLDataBaseTool):
                # Add our own custom description so our agent knows when to query this table.
                original_description = tool.description
                tool.description = ''
                if len(descriptions) > 0:
                    tool.description += f'Use this tool if you need data about {" OR ".join(descriptions)}.\n'
                tool.description += 'Use the conversation context to decide which table to query.\n'
                if len(tables_list) > 0:
                    f'These are the available tables: {",".join(tables_list)}.\n'
                tool.description += (
                    'ALWAYS consider these special cases:\n'
                    ' - For TIMESTAMP type columns, make sure you include the time portion in your query (e.g. WHERE date_column = "2020-01-01 12:00:00")\n'
                    'Here are the rest of the instructions:\n'
                    f'{original_description}'
                )
                sql_database_tools[i] = tool
        return sql_database_tools

    def _make_retrieval_tools(self, skill: db.Skills, llm, embedding_model):
        """
        creates advanced retrieval tool i.e. RAG
        """
        params = skill.params
        config = params.get('config', {})
        if 'llm' not in config:
            # Set LLM if not explicitly provided in configs.
            config['llm'] = llm
        tool = dict(
            name=params.get('name', skill.name),
            source=params.get('source', None),
            config=config,
            description=f'You must use this tool to get more context or information '
                        f'to answer a question about {params["description"]}. '
                        f'The input should be the exact question the user is asking.',
            type=skill.type
        )
        pred_args = {}
        pred_args['llm'] = llm

        from .retrieval_tool import build_retrieval_tool
        return build_retrieval_tool(tool, pred_args, skill)

    def _get_rag_query_function(self, skill: db.Skills):
        session_controller = self.get_command_executor().session

        def _answer_question(question: str) -> str:
            knowledge_base_name = skill.params['source']

            # make select in KB table
            query = Select(
                targets=[Star()],
                where=BinaryOperation(op='=', args=[
                    Identifier(TableField.CONTENT.value), Constant(question)
                ]),
                limit=Constant(_DEFAULT_TOP_K_SIMILARITY_SEARCH),
            )
            kb_table = session_controller.kb_controller.get_table(knowledge_base_name, skill.project_id)

            res = kb_table.select_query(query)
            return '\n'.join(res.content)

        return _answer_question

    def _make_knowledge_base_tools(self, skill: db.Skills) -> dict:
        # To prevent dependency on Langchain unless an actual tool uses it.
        description = skill.params.get('description', '')

        logger.warning("This skill is deprecated and will be removed in the future. "
                       "Please use `retrieval` skill instead ")

        return dict(
            name='Knowledge Base Retrieval',
            func=self._get_rag_query_function(skill),
            description=f'Use this tool to get more context or information to answer a question about {description}. The input should be the exact question the user is asking.',
            type=skill.type
        )

    def get_tools_from_skills(self, skills_data: List[SkillData], llm: BaseChatModel, embedding_model: Embeddings) -> dict:
        """Creates function for skill and metadata (name, description)

        Args:
            skills_data (List[SkillData]): Skills to make a tool from
            llm (BaseChatModel): LLM which will be used by skills
            embedding_model (Embeddings): this model is used by retrieval skill

        Returns:
            dict: with keys: name, description, func
        """

        # group skills by type
        skills_group = defaultdict(list)
        for skill in skills_data:
            try:
                skill_type = SkillType(skill.type)
            except ValueError:
                raise NotImplementedError(
                    f'skill of type {skill.type} is not supported as a tool, supported types are: {list(SkillType._member_names_)}')

            if skill_type == SkillType.TEXT2SQL_LEGACY:
                skill_type = SkillType.TEXT2SQL
            skills_group[skill_type].append(skill)

        tools = {}
        for skill_type, skills in skills_group.items():
            if skill_type == SkillType.TEXT2SQL:
                tools[skill_type] = self._make_text_to_sql_tools(skills, llm)
            elif skill_type == SkillType.KNOWLEDGE_BASE:
                tools[skill_type] = [
                    self._make_knowledge_base_tools(skill)
                    for skill in skills
                ]
            elif skill_type == SkillType.RETRIEVAL:
                tools[skill_type] = [
                    self._make_retrieval_tools(skill, llm, embedding_model)
                    for skill in skills
                ]
        return tools


skill_tool = SkillToolController()



import importlib

from pandas import DataFrame

import mindsdb.interfaces.storage.db as db
from mindsdb.interfaces.storage.model_fs import ModelStorage, HandlerStorage
from mindsdb.integrations.libs.ml_handler_process.handlers_cacher import handlers_cacher
from mindsdb.utilities.functions import mark_process


@mark_process(name='learn')
def predict_process(integration_id: int, predictor_record: db.Predictor, args: dict,
                    module_path: str, ml_engine_name: str, dataframe: DataFrame) -> DataFrame:
    module = importlib.import_module(module_path)

    if predictor_record.id not in handlers_cacher:
        handlerStorage = HandlerStorage(integration_id)
        modelStorage = ModelStorage(predictor_record.id)
        ml_handler = module.Handler(
            engine_storage=handlerStorage,
            model_storage=modelStorage,
        )
        handlers_cacher[predictor_record.id] = ml_handler
    else:
        ml_handler = handlers_cacher[predictor_record.id]

    if ml_engine_name == 'lightwood':
        args['code'] = predictor_record.code
        args['target'] = predictor_record.to_predict[0]
        args['dtype_dict'] = predictor_record.dtype_dict
        args['learn_args'] = predictor_record.learn_args

    predictions = ml_handler.predict(dataframe, args)
    ml_handler.close()
    return predictions



from typing import Text, Dict, Any
from google.api_core.exceptions import BadRequest
from sqlalchemy_bigquery.base import BigQueryDialect
from google.cloud.bigquery import Client, QueryJobConfig

from mindsdb.utilities import log
from mindsdb_sql_parser.ast.base import ASTNode
from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.utilities.render.sqlalchemy_render import SqlalchemyRender
from mindsdb.integrations.utilities.handlers.auth_utilities import GoogleServiceAccountOAuth2Manager
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)

logger = log.getLogger(__name__)


class BigQueryHandler(DatabaseHandler):
    """
    This handler handles connection and execution of Google BigQuery statements.
    """
    name = "bigquery"

    def __init__(self, name: Text, connection_data: Dict, **kwargs: Any):
        super().__init__(name)
        self.connection_data = connection_data
        self.client = None
        self.is_connected = False

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def connect(self):
        """
        Establishes a connection to a BigQuery warehouse.

        Raises:
            ValueError: If the required connection parameters are not provided or if the credentials cannot be parsed.
            mindsdb.integrations.utilities.handlers.auth_utilities.exceptions.NoCredentialsException: If none of the required forms of credentials are provided.
            mindsdb.integrations.utilities.handlers.auth_utilities.exceptions.AuthException: If authentication fails.

        Returns:
            google.cloud.bigquery.client.Client: The client object for the BigQuery connection.
        """
        if self.is_connected is True:
            return self.connection

        # Mandatory connection parameters
        if not all(key in self.connection_data for key in ['project_id', 'dataset']):
            raise ValueError('Required parameters (project_id, dataset) must be provided.')

        google_sa_oauth2_manager = GoogleServiceAccountOAuth2Manager(
            credentials_file=self.connection_data.get('service_account_keys'),
            credentials_json=self.connection_data.get('service_account_json')
        )
        credentials = google_sa_oauth2_manager.get_oauth2_credentials()

        client = Client(
            project=self.connection_data["project_id"],
            credentials=credentials
        )
        self.is_connected = True
        self.connection = client
        return self.connection

    def disconnect(self):
        """
        Closes the connection to the BigQuery warehouse if it's currently open.
        """
        if self.is_connected is False:
            return
        self.connection.close()
        self.is_connected = False

    def check_connection(self) -> StatusResponse:
        """
        Checks the status of the connection to the BigQuery warehouse.

        Returns:
            StatusResponse: An object containing the success status and an error message if an error occurs.
        """
        response = StatusResponse(False)

        try:
            connection = self.connect()
            connection.query('SELECT 1;')

            # Check if the dataset exists
            connection.get_dataset(self.connection_data['dataset'])

            response.success = True
        except (BadRequest, ValueError) as e:
            logger.error(f'Error connecting to BigQuery {self.connection_data["project_id"]}, {e}!')
            response.error_message = e

        if response.success is False and self.is_connected is True:
            self.is_connected = False

        return response

    def native_query(self, query: str) -> Response:
        """
        Executes a SQL query on the BigQuery warehouse and returns the result.

        Args:
            query (str): The SQL query to be executed.

        Returns:
            Response: A response object containing the result of the query or an error message.
        """
        connection = self.connect()
        try:
            job_config = QueryJobConfig(default_dataset=f"{self.connection_data['project_id']}.{self.connection_data['dataset']}")
            query = connection.query(query, job_config=job_config)
            result = query.to_dataframe()
            if not result.empty:
                response = Response(
                    RESPONSE_TYPE.TABLE,
                    result
                )
            else:
                response = Response(RESPONSE_TYPE.OK)
        except Exception as e:
            logger.error(f'Error running query: {query} on {self.connection_data["project_id"]}!')
            response = Response(
                RESPONSE_TYPE.ERROR,
                error_message=str(e)
            )
        return response

    def query(self, query: ASTNode) -> Response:
        """
        Executes a SQL query represented by an ASTNode and retrieves the data.

        Args:
            query (ASTNode): An ASTNode representing the SQL query to be executed.

        Returns:
            Response: The response from the `native_query` method, containing the result of the SQL query execution.
        """
        renderer = SqlalchemyRender(BigQueryDialect)
        query_str = renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)

    def get_tables(self) -> Response:
        """
        Retrieves a list of all non-system tables and views in the configured dataset of the BigQuery warehouse.

        Returns:
            Response: A response object containing the list of tables and views, formatted as per the `Response` class.
        """
        query = f"""
            SELECT table_name, table_schema, table_type
            FROM `{self.connection_data['project_id']}.{self.connection_data['dataset']}.INFORMATION_SCHEMA.TABLES`
            WHERE table_type IN ('BASE TABLE', 'VIEW')
        """
        result = self.native_query(query)
        return result

    def get_columns(self, table_name) -> Response:
        """
        Retrieves column details for a specified table in the configured dataset of the BigQuery warehouse.

        Args:
            table_name (str): The name of the table for which to retrieve column information.

        Returns:
            Response: A response object containing the column details, formatted as per the `Response` class.
        Raises:
            ValueError: If the 'table_name' is not a valid string.
        """
        query = f"""
            SELECT column_name AS Field, data_type as Type
            FROM `{self.connection_data['project_id']}.{self.connection_data['dataset']}.INFORMATION_SCHEMA.COLUMNS`
            WHERE table_name = '{table_name}'
        """
        result = self.native_query(query)
        return result



"""predictor-hostname

Revision ID: b0382f5be48d
Revises: 011e6f2dd9c2
Create Date: 2023-08-29 17:19:55.372394

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'b0382f5be48d'
down_revision = '011e6f2dd9c2'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.add_column(sa.Column('hostname', sa.String(), nullable=True))


def downgrade():
    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.drop_column('hostname')



import json
from concurrent.futures import as_completed, TimeoutError
from typing import Dict, Iterable, List, Optional
from uuid import uuid4
import re
import numpy as np
import pandas as pd

from langchain.agents import AgentExecutor
from langchain.agents.initialize import initialize_agent
from langchain.chains.conversation.memory import ConversationSummaryBufferMemory
from langchain_community.chat_models import (
    ChatAnyscale,
    ChatLiteLLM,
    ChatOllama)
from langchain_core.agents import AgentAction, AgentStep

from langchain_nvidia_ai_endpoints import ChatNVIDIA
from langchain_core.messages.base import BaseMessage
from langchain_core.prompts import PromptTemplate
from langchain_core.tools import Tool

from mindsdb.integrations.handlers.openai_handler.constants import (
    CHAT_MODELS as OPEN_AI_CHAT_MODELS,
)
from mindsdb.integrations.libs.llm.utils import get_llm_config
from mindsdb.integrations.utilities.handler_utils import get_api_key
from mindsdb.integrations.utilities.rag.settings import DEFAULT_RAG_PROMPT_TEMPLATE
from mindsdb.utilities import log
from mindsdb.utilities.context_executor import ContextThreadPoolExecutor
from mindsdb.interfaces.storage import db
from mindsdb.utilities.context import context as ctx

from .mindsdb_chat_model import ChatMindsdb
from .callback_handlers import LogCallbackHandler, ContextCaptureCallback
from .langfuse_callback_handler import LangfuseCallbackHandler, get_skills
from .safe_output_parser import SafeOutputParser

from .constants import (
    DEFAULT_AGENT_TIMEOUT_SECONDS,
    DEFAULT_AGENT_TYPE,
    DEFAULT_EMBEDDINGS_MODEL_PROVIDER,
    DEFAULT_MAX_ITERATIONS,
    DEFAULT_MAX_TOKENS,
    DEFAULT_TIKTOKEN_MODEL_NAME,
    SUPPORTED_PROVIDERS,
    ANTHROPIC_CHAT_MODELS,
    OLLAMA_CHAT_MODELS,
    NVIDIA_NIM_CHAT_MODELS,
    USER_COLUMN,
    ASSISTANT_COLUMN,
    CONTEXT_COLUMN, TRACE_ID_COLUMN
)
from mindsdb.interfaces.skills.skill_tool import skill_tool, SkillData
from langchain_anthropic import ChatAnthropic
from langchain_core.messages import SystemMessage
from langchain_openai import ChatOpenAI

from mindsdb.utilities.langfuse import LangfuseClientWrapper

_PARSING_ERROR_PREFIXES = [
    "An output parsing error occurred",
    "Could not parse LLM output",
]

logger = log.getLogger(__name__)


def get_llm_provider(args: Dict) -> str:
    # If provider is explicitly specified, use that
    if "provider" in args:
        return args["provider"]

    # Check for known model names from other providers first
    if args["model_name"] in ANTHROPIC_CHAT_MODELS:
        return "anthropic"
    if args["model_name"] in OPEN_AI_CHAT_MODELS:
        return "openai"
    if args["model_name"] in OLLAMA_CHAT_MODELS:
        return "ollama"
    if args["model_name"] in NVIDIA_NIM_CHAT_MODELS:
        return "nvidia_nim"

    # For vLLM, require explicit provider specification
    raise ValueError("Invalid model name. Please define a supported llm provider")


def get_embedding_model_provider(args: Dict) -> str:
    """Get the embedding model provider from args.

    For VLLM, this will use our custom VLLMEmbeddings class from langchain_embedding_handler.
    """
    # Check for explicit embedding model provider
    if "embedding_model_provider" in args:
        provider = args["embedding_model_provider"]
        if provider == 'vllm':
            if not (args.get('openai_api_base') and args.get('model')):
                raise ValueError(
                    "VLLM embeddings configuration error:\n"
                    "- Missing required parameters: 'openai_api_base' and/or 'model'\n"
                    "- Example: openai_api_base='http://localhost:8003/v1', model='your-model-name'"
                )
            logger.info("Using custom VLLMEmbeddings class")
            return 'vllm'
        return provider

    # Check if LLM provider is vLLM
    llm_provider = args.get('provider', DEFAULT_EMBEDDINGS_MODEL_PROVIDER)
    if llm_provider == 'vllm':
        if not (args.get('openai_api_base') and args.get('model')):
            raise ValueError(
                "VLLM embeddings configuration error:\n"
                "- Missing required parameters: 'openai_api_base' and/or 'model'\n"
                "- When using VLLM as LLM provider, you must specify the embeddings server location and model\n"
                "- Example: openai_api_base='http://localhost:8003/v1', model='your-model-name'"
            )
        logger.info("Using custom VLLMEmbeddings class")
        return 'vllm'

    # Default to LLM provider
    return llm_provider


def get_chat_model_params(args: Dict) -> Dict:
    model_config = args.copy()
    # Include API keys.
    model_config["api_keys"] = {
        p: get_api_key(p, model_config, None, strict=False) for p in SUPPORTED_PROVIDERS
    }
    llm_config = get_llm_config(
        args.get("provider", get_llm_provider(args)), model_config
    )
    config_dict = llm_config.model_dump()
    config_dict = {k: v for k, v in config_dict.items() if v is not None}
    return config_dict


def create_chat_model(args: Dict):
    model_kwargs = get_chat_model_params(args)

    if args["provider"] == "anthropic":
        return ChatAnthropic(**model_kwargs)
    if args["provider"] == "openai" or args["provider"] == "vllm":
        chat_open_ai = ChatOpenAI(**model_kwargs)
        # Some newer GPT models (e.g. gpt-4o when released) don't have token counting support yet.
        # By setting this manually in ChatOpenAI, we count tokens like compatible GPT models.
        try:
            chat_open_ai.get_num_tokens_from_messages([])
        except NotImplementedError:
            chat_open_ai.tiktoken_model_name = DEFAULT_TIKTOKEN_MODEL_NAME
        return chat_open_ai
    if args["provider"] == "anyscale":
        return ChatAnyscale(**model_kwargs)
    if args["provider"] == "litellm":
        return ChatLiteLLM(**model_kwargs)
    if args["provider"] == "ollama":
        return ChatOllama(**model_kwargs)
    if args["provider"] == "nvidia_nim":
        return ChatNVIDIA(**model_kwargs)
    if args["provider"] == "mindsdb":
        return ChatMindsdb(**model_kwargs)
    raise ValueError(f'Unknown provider: {args["provider"]}')


def prepare_prompts(df, base_template, input_variables, user_column=USER_COLUMN):
    empty_prompt_ids = np.where(df[input_variables].isna().all(axis=1).values)[0]
    base_template = base_template.replace('{{', '{').replace('}}', '}')
    prompts = []

    for i, row in df.iterrows():
        if i not in empty_prompt_ids:
            prompt = PromptTemplate(input_variables=input_variables, template=base_template)
            kwargs = {col: row[col] if row[col] is not None else '' for col in input_variables}
            prompts.append(prompt.format(**kwargs))
        elif row.get(user_column):
            prompts.append(row[user_column])

    return prompts, empty_prompt_ids


def prepare_callbacks(self, args):
    context_callback = ContextCaptureCallback()
    callbacks = self._get_agent_callbacks(args)
    callbacks.append(context_callback)
    return callbacks, context_callback


def handle_agent_error(e):
    error_message = f"An error occurred during agent execution: {str(e)}"
    logger.error(error_message, exc_info=True)
    return error_message


def process_chunk(chunk):
    if isinstance(chunk, dict):
        return {k: process_chunk(v) for k, v in chunk.items()}
    elif isinstance(chunk, list):
        return [process_chunk(item) for item in chunk]
    elif isinstance(chunk, (str, int, float, bool, type(None))):
        return chunk
    else:
        return str(chunk)


class LangchainAgent:

    def __init__(self, agent: db.Agents, model: dict = None):

        self.agent = agent
        self.model = model

        self.run_completion_span: Optional[object] = None
        self.llm: Optional[object] = None
        self.embedding_model: Optional[object] = None

        self.log_callback_handler: Optional[object] = None
        self.langfuse_callback_handler: Optional[object] = None  # native langfuse callback handler
        self.mdb_langfuse_callback_handler: Optional[object] = None  # custom (see langfuse_callback_handler.py)

        self.langfuse_client_wrapper = LangfuseClientWrapper()
        self.args = self._initialize_args()

        # Back compatibility for old models
        self.provider = self.args.get("provider", get_llm_provider(self.args))

    def _initialize_args(self) -> dict:
        """Initialize the arguments based on the agent's parameters."""
        args = self.agent.params.copy()
        args["model_name"] = self.agent.model_name
        args["provider"] = self.agent.provider
        args["embedding_model_provider"] = args.get(
            "embedding_model", get_embedding_model_provider(args)
        )

        # agent is using current langchain model
        if self.agent.provider == "mindsdb":
            args["model_name"] = self.agent.model_name

            # get prompt
            prompt_template = (
                self.model["problem_definition"].get("using", {}).get("prompt_template")
            )
            if prompt_template is not None:
                # only update prompt_template if it is set on the model
                args["prompt_template"] = prompt_template

        if args.get("prompt_template") is None:
            if args.get("mode") == "retrieval":
                args["prompt_template"] = DEFAULT_RAG_PROMPT_TEMPLATE
            else:
                raise ValueError(
                    "Please provide a `prompt_template` or set `mode=retrieval`"
                )

        return args

    def get_metadata(self) -> Dict:
        return {
            'provider': self.provider,
            'model_name': self.args["model_name"],
            'embedding_model_provider': self.args.get('embedding_model_provider',
                                                      get_embedding_model_provider(self.args)),
            'skills': get_skills(self.agent),
            'user_id': ctx.user_id,
            'session_id': ctx.session_id,
            'company_id': ctx.company_id,
            'user_class': ctx.user_class,
            'email_confirmed': ctx.email_confirmed
        }

    def get_tags(self) -> List:
        return [
            self.provider,
        ]

    def get_completion(self, messages, stream: bool = False):

        # Get metadata and tags to be used in the trace
        metadata = self.get_metadata()
        tags = self.get_tags()

        # Set up trace for the API completion in Langfuse
        self.langfuse_client_wrapper.setup_trace(
            name='api-completion',
            input=messages,
            tags=tags,
            metadata=metadata,
            user_id=ctx.user_id,
            session_id=ctx.session_id,
        )

        # Set up trace for the run completion in Langfuse
        self.run_completion_span = self.langfuse_client_wrapper.start_span(
            name='run-completion',
            input=messages)

        if stream:
            return self._get_completion_stream(messages)

        args = self.args

        df = pd.DataFrame(messages)

        # Back compatibility for old models
        self.provider = args.get("provider", get_llm_provider(args))

        df = df.reset_index(drop=True)
        agent = self.create_agent(df, args)
        # Use last message as prompt, remove other questions.
        user_column = args.get("user_column", USER_COLUMN)
        df.iloc[:-1, df.columns.get_loc(user_column)] = None
        response = self.run_agent(df, agent, args)

        # End the run completion span and update the metadata with tool usage
        self.langfuse_client_wrapper.end_span(span=self.run_completion_span, output=response)

        return response

    def _get_completion_stream(self, messages: List[dict]) -> Iterable[Dict]:
        """Gets a completion as a stream of chunks from given messages.

        Args:
            messages (List[dict]): Messages to get completion chunks for

        Returns:
            chunks (Iterable[object]): Completion chunks
        """

        args = self.args

        df = pd.DataFrame(messages)

        self.embedding_model_provider = args.get('embedding_model_provider', get_embedding_model_provider(args))
        # Back compatibility for old models
        self.provider = args.get("provider", get_llm_provider(args))

        df = df.reset_index(drop=True)
        agent = self.create_agent(df, args)
        # Use last message as prompt, remove other questions.
        user_column = args.get("user_column", USER_COLUMN)
        df.iloc[:-1, df.columns.get_loc(user_column)] = None
        return self.stream_agent(df, agent, args)

    def create_agent(self, df: pd.DataFrame, args: Dict = None) -> AgentExecutor:
        # Set up tools.
        llm = create_chat_model(args)
        self.llm = llm

        # Don't set embedding model for retrieval mode - let the knowledge base handle it
        if args.get("mode") == "retrieval":
            self.args.pop("mode")

        tools = self._langchain_tools_from_skills(llm)

        # Prefer prediction prompt template over original if provided.
        prompt_template = args["prompt_template"]

        # Set up memory.
        memory = ConversationSummaryBufferMemory(
            llm=llm,
            input_key="input",
            output_key="output",
            max_token_limit=args.get("max_tokens", DEFAULT_MAX_TOKENS),
            memory_key="chat_history",
        )

        memory.chat_memory.messages.insert(0, SystemMessage(content=prompt_template))
        # User - Assistant conversation. All except the last message.
        user_column = args.get("user_column", USER_COLUMN)
        assistant_column = args.get("assistant_column", ASSISTANT_COLUMN)
        for row in df[:-1].to_dict("records"):
            question = row[user_column]
            answer = row[assistant_column]
            if isinstance(question, str) and len(question) > 0:
                memory.chat_memory.add_user_message(question)
            if isinstance(answer, str) and len(answer) > 0:
                memory.chat_memory.add_ai_message(answer)

        agent_type = args.get("agent_type", DEFAULT_AGENT_TYPE)
        agent_executor = initialize_agent(
            tools,
            llm,
            agent=agent_type,
            # Use custom output parser to handle flaky LLMs that don't ALWAYS conform to output format.
            agent_kwargs={"output_parser": SafeOutputParser()},
            # Calls the agent’s LLM Chain one final time to generate a final answer based on the previous steps
            early_stopping_method="generate",
            handle_parsing_errors=self._handle_parsing_errors,
            # Timeout per agent invocation.
            max_execution_time=args.get(
                "timeout_seconds",
                args.get("timeout_seconds", DEFAULT_AGENT_TIMEOUT_SECONDS),
            ),
            max_iterations=args.get(
                "max_iterations", args.get("max_iterations", DEFAULT_MAX_ITERATIONS)
            ),
            memory=memory,
            verbose=args.get("verbose", args.get("verbose", True)),
        )
        return agent_executor

    def _langchain_tools_from_skills(self, llm):
        # Makes Langchain compatible tools from a skill
        skills_data = [
            SkillData(
                name=rel.skill.name,
                type=rel.skill.type,
                params=rel.skill.params,
                project_id=rel.skill.project_id,
                agent_tables_list=(rel.parameters or {}).get('tables')
            )
            for rel in self.agent.skills_relationships
        ]

        tools_groups = skill_tool.get_tools_from_skills(skills_data, llm, self.embedding_model)

        all_tools = []
        for skill_type, tools in tools_groups.items():
            for tool in tools:
                if isinstance(tool, dict):
                    tool = Tool(
                        name=tool["name"],
                        func=tool["func"],
                        description=tool["description"],
                    )
                all_tools.append(tool)
        return all_tools

    def _get_agent_callbacks(self, args: Dict) -> List:
        all_callbacks = []

        if self.log_callback_handler is None:
            self.log_callback_handler = LogCallbackHandler(logger)

        all_callbacks.append(self.log_callback_handler)

        if self.langfuse_client_wrapper.trace is None:
            # Get metadata and tags to be used in the trace
            metadata = self.get_metadata()
            tags = self.get_tags()

            trace_name = "NativeTrace-MindsDB-AgentExecutor"

            # Set up trace for the API completion in Langfuse
            self.langfuse_client_wrapper.setup_trace(
                name=trace_name,
                tags=tags,
                metadata=metadata,
                user_id=ctx.user_id,
                session_id=ctx.session_id,
            )

        if self.langfuse_callback_handler is None:
            self.langfuse_callback_handler = self.langfuse_client_wrapper.get_langchain_handler()

        # custom tracer
        if self.mdb_langfuse_callback_handler is None:
            trace_id = self.langfuse_client_wrapper.get_trace_id()

            span_id = None
            if self.run_completion_span is not None:
                span_id = self.run_completion_span.id

            observation_id = args.get("observation_id", span_id or uuid4().hex)

            self.mdb_langfuse_callback_handler = LangfuseCallbackHandler(
                langfuse=self.langfuse_client_wrapper.client,
                trace_id=trace_id,
                observation_id=observation_id,
            )

        # obs: we may want to unify these; native langfuse handler provides details as a tree on a sub-step of the overarching custom one  # noqa
        if self.langfuse_callback_handler is not None:
            all_callbacks.append(self.langfuse_callback_handler)

        if self.mdb_langfuse_callback_handler:
            all_callbacks.append(self.mdb_langfuse_callback_handler)

        return all_callbacks

    def _handle_parsing_errors(self, error: Exception) -> str:
        response = str(error)
        for p in _PARSING_ERROR_PREFIXES:
            if response.startswith(p):
                # As a somewhat dirty workaround, we accept the output formatted incorrectly and use it as a response.
                #
                # Ideally, in the future, we would write a parser that is more robust and flexible than the one Langchain uses.
                # Response is wrapped in ``
                logger.info("Handling parsing error, salvaging response...")
                response_output = response.split("`")
                if len(response_output) >= 2:
                    response = response_output[-2]

                # Wrap response in Langchain conversational react format.
                langchain_react_formatted_response = f"""Thought: Do I need to use a tool? No
AI: {response}"""
                return langchain_react_formatted_response
        return f"Agent failed with error:\n{str(error)}..."

    def run_agent(self, df: pd.DataFrame, agent: AgentExecutor, args: Dict) -> pd.DataFrame:
        base_template = args.get('prompt_template', args['prompt_template'])
        return_context = args.get('return_context', True)
        input_variables = re.findall(r"{{(.*?)}}", base_template)

        prompts, empty_prompt_ids = prepare_prompts(df, base_template, input_variables,
                                                    args.get('user_column', USER_COLUMN))

        def _invoke_agent_executor_with_prompt(agent_executor, prompt):
            if not prompt:
                return {CONTEXT_COLUMN: [], ASSISTANT_COLUMN: ""}
            try:
                callbacks, context_callback = prepare_callbacks(self, args)
                result = agent_executor.invoke(prompt, config={'callbacks': callbacks})
                captured_context = context_callback.get_contexts()
                output = result['output'] if isinstance(result, dict) and 'output' in result else str(result)
                return {CONTEXT_COLUMN: captured_context, ASSISTANT_COLUMN: output}
            except Exception as e:
                return {CONTEXT_COLUMN: [], ASSISTANT_COLUMN: handle_agent_error(e)}

        completions = []
        contexts = []

        max_workers = args.get("max_workers", None)
        agent_timeout_seconds = args.get("timeout", DEFAULT_AGENT_TIMEOUT_SECONDS)

        with ContextThreadPoolExecutor(max_workers=max_workers) as executor:
            futures = [
                executor.submit(_invoke_agent_executor_with_prompt, agent, prompt)
                for prompt in prompts
            ]
            try:
                for future in as_completed(futures, timeout=agent_timeout_seconds):
                    result = future.result()
                    if result is None:
                        result = {
                            CONTEXT_COLUMN: [],
                            ASSISTANT_COLUMN: "No response generated",
                        }

                    completions.append(result[ASSISTANT_COLUMN])
                    contexts.append(result[CONTEXT_COLUMN])
            except TimeoutError:
                timeout_message = "I'm sorry! I couldn't come up with a response in time. Please try again."
                logger.warning(
                    f"Agent execution timed out after {agent_timeout_seconds} seconds"
                )
                for _ in range(len(futures) - len(completions)):
                    completions.append(timeout_message)
                    contexts.append([])

        # Add null completion for empty prompts
        for i in sorted(empty_prompt_ids)[:-1]:
            completions.insert(i, None)
            contexts.insert(i, [])

        # Create DataFrame with completions and context if required
        pred_df = pd.DataFrame(
            {
                ASSISTANT_COLUMN: completions,
                CONTEXT_COLUMN: [
                    json.dumps(ctx) for ctx in contexts
                ],  # Serialize context to JSON string
                TRACE_ID_COLUMN: self.langfuse_client_wrapper.get_trace_id()
            }
        )

        if not return_context:
            pred_df = pred_df.drop(columns=[CONTEXT_COLUMN])

        return pred_df

    def add_chunk_metadata(self, chunk: Dict) -> Dict:
        logger.debug(f'Adding metadata to chunk: {chunk}')
        logger.debug(f'Trace ID: {self.langfuse_client_wrapper.get_trace_id()}')
        chunk["trace_id"] = self.langfuse_client_wrapper.get_trace_id()
        return chunk

    def stream_agent(self, df: pd.DataFrame, agent_executor: AgentExecutor, args: Dict) -> Iterable[Dict]:
        base_template = args.get('prompt_template', args['prompt_template'])
        input_variables = re.findall(r"{{(.*?)}}", base_template)
        return_context = args.get('return_context', True)

        prompts, _ = prepare_prompts(df, base_template, input_variables, args.get('user_column', USER_COLUMN))

        callbacks, context_callback = prepare_callbacks(self, args)

        yield self.add_chunk_metadata({"type": "start", "prompt": prompts[0]})

        if not hasattr(agent_executor, 'stream') or not callable(agent_executor.stream):
            raise AttributeError("The agent_executor does not have a 'stream' method")

        stream_iterator = agent_executor.stream(prompts[0],
                                                config={'callbacks': callbacks})

        if not hasattr(stream_iterator, '__iter__'):
            raise TypeError("The stream method did not return an iterable")

        for chunk in stream_iterator:
            logger.debug(f'Processing streaming chunk {chunk}')
            processed_chunk = self.process_chunk(chunk)
            logger.info(f'Processed chunk: {processed_chunk}')
            yield self.add_chunk_metadata(processed_chunk)

        if return_context:
            # Yield context if required
            captured_context = context_callback.get_contexts()
            if captured_context:
                yield {"type": "context", "content": captured_context}

        if self.log_callback_handler.generated_sql:
            # Yield generated SQL if available
            yield self.add_chunk_metadata({"type": "sql", "content": self.log_callback_handler.generated_sql})

        # End the run completion span and update the metadata with tool usage
        self.langfuse_client_wrapper.end_span_stream(span=self.run_completion_span)

    @staticmethod
    def process_chunk(chunk):
        if isinstance(chunk, dict):
            return {k: LangchainAgent.process_chunk(v) for k, v in chunk.items()}
        if isinstance(chunk, list):
            return [LangchainAgent.process_chunk(item) for item in chunk]
        if isinstance(chunk, AgentAction):
            # Format agent actions properly for streaming.
            return {
                'tool': LangchainAgent.process_chunk(chunk.tool),
                'tool_input': LangchainAgent.process_chunk(chunk.tool_input),
                'log': LangchainAgent.process_chunk(chunk.log)
            }
        if isinstance(chunk, AgentStep):
            # Format agent steps properly for streaming.
            return {
                'action': LangchainAgent.process_chunk(chunk.action),
                'observation': LangchainAgent.process_chunk(chunk.observation) if chunk.observation else ''
            }
        if issubclass(chunk.__class__, BaseMessage):
            # Extract content from message subclasses properly for streaming.
            return {
                'content': chunk.content
            }
        if isinstance(chunk, (str, int, float, bool, type(None))):
            return chunk
        return str(chunk)



import requests
from urllib.parse import urljoin


class Lightdash:

    def __init__(self, url: str, api_key: str) -> None:
        self.base_url = urljoin(url, "/api/v1/")
        self.api_key = api_key

    def _request(self, method: str, relative_endpoint: str, data=None):
        kwargs = {
            "method": method,
            "url": urljoin(self.base_url, relative_endpoint),
            "headers": {
                "Authorization": "ApiKey " + self.api_key,
            }
        }
        if data is not None:
            kwargs["data"] = data
        return requests.request(**kwargs)

    def is_connected(self) -> bool:
        if self.get_user() is not None:
            return True
        return False

    def get_user(self):
        """
        Get user's details
        Return format:
            {'userUuid': '831b6c26-bdc2-4a56-9818-fd8ebaa406ac',
              'email': 'test@test.com',
              'firstName': 'Test',
              'lastName': 'User',
              'organizationUuid': 'd00805a0-b0b4-400d-a136-66f620493f11',
              'organizationName': 'testing-comp',
              'organizationCreatedAt': '2023-10-20T17:46:10.005Z',
              'isTrackingAnonymized': False,
              'isMarketingOptedIn': True,
              'isSetupComplete': True,
              'role': 'admin',
              'isActive': True,
              'abilityRules': [{'action': 'view',
                'subject': 'OrganizationMemberProfile',
                'conditions': {'organizationUuid': 'd00805a0-b0b4-400d-a136-66f620493f11'}},
               {'action': ...
        """
        resp = self._request("get", "user")
        if resp.ok:
            return resp.json()["results"]
        return {}

    def get_org(self):
        """
        Get user's organization details
        Return format:
            {'organizationUuid': 'd00805a0-b0b4-400d-a136-66f620493f11',
              'name': 'testing-comp',
              'defaultProjectUuid': 'string',
              'chartColors': ['string'],
              'needsProject': False}
        """
        resp = self._request("get", "org")
        if resp.ok:
            return resp.json()["results"]
        return {}

    def get_projects(self):
        """
        Get user organization's pojects' details
        Return format:
            [{'name': 'Jaffle shop',
              'projectUuid': '95dfda3b-02e2-4708-a014-5967966020f3',
              'type': 'DEFAULT'}]
        """
        resp = self._request("get", "org/projects")
        if resp.ok:
            return resp.json()["results"]
        return []

    def get_org_members(self):
        """
        Get all the members in user's organization
        Return format:
            [{'userUuid': '831b6c26-bdc2-4a56-9818-fd8ebaa406ac',
              'firstName': 'TestName',
              'lastName': 'TestLastName',
              'email': 'test@test.com',
              'organizationUuid': 'd00805a0-b0b4-400d-a136-66f620493f11',
              'role': 'admin',
              'isActive': True,
              'isInviteExpired': False}]
        """
        resp = self._request("get", "org/users")
        if resp.ok:
            return resp.json()["results"]
        return []

    def get_project(self, project_uuid: str):
        """
        Get details of a project in user's organization
        Return format:
            { "dbtVersion": "v1.4",
              "copiedFromProjectUuid": "string",
              "pinnedListUuid": "string",
              "warehouseConnection": {
                "role": "string",
                "type": "snowflake",
                "account": "string",
                "database": "string",
                "warehouse": "string",
                "schema": "string",
                "threads": 0,
                "clientSessionKeepAlive": true,
                "queryTag": "string",
                "accessUrl": "string",
                "startOfWeek": 0 },
              "dbtConnection": {
                "type": "dbt",
                "target": "string",
                "environment": [ {
                    "value": "string",
                    "key": "string" } ],
                "profiles_dir": "string",
                "project_dir": "string" },
              "type": "DEFAULT",
              "name": "string",
              "projectUuid": "string",
              "organizationUuid": "string" }
        """
        resp = self._request("get", f"projects/{project_uuid}")
        if resp.ok:
            return resp.json()["results"]
        return {}

    def get_charts_in_project(self, project_uuid: str):
        """
        List all charts in a project
        Return format:
            [{"name": "string",
              "organizationUuid": "string",
              "uuid": "string",
              "description": "string",
              "projectUuid": "string",
              "spaceUuid": "string",
              "pinnedListUuid": "string",
              "spaceName": "string",
              "dashboardUuid": "string",
              "dashboardName": "string",
              "chartType": "string"}]
        """
        resp = self._request("get", f"projects/{project_uuid}/charts")
        if resp.ok:
            return resp.json()["results"]
        return []

    def get_spaces_in_project(self, project_uuid: str):
        """
        List all spaces in a project
        Return format:
            [{"name": "string",
              "organizationUuid": "string",
              "uuid": "string",
              "projectUuid": "string",
              "pinnedListUuid": "string",
              "pinnedListOrder": 0,
              "isPrivate": true,
              "dashboardCount": 0,
              "chartCount": 0,
              "access": [
                  "string" ]}]
        """
        resp = self._request("get", f"projects/{project_uuid}/spaces")
        if resp.ok:
            return resp.json()["results"]
        return []

    def get_project_access_list(self, project_uuid: str):
        """
        Get access list for a project. This is a list of users that have been explictly granted access to the project. There may be other users that have access to the project via their organization membership
        Return format:
            [{"lastName": "string",
              "firstName": "string",
              "email": "string",
              "role": "viewer",
              "projectUuid": "string",
              "userUuid": "string" }]
        """
        resp = self._request("get", f"projects/{project_uuid}/access")
        if resp.ok:
            return resp.json()["results"]
        return []

    def get_validation_results(self, project_uuid: str):
        """
        Get validation results for a project. This will return the results of the latest validation job
        Return format:
            [{"source": "chart",
              "spaceUuid": "string",
              "projectUuid": "string",
              "errorType": "chart",
              "error": "string",
              "name": "string",
              "createdAt": "2019-08-24T14:15:22Z",
              "validationId": 0,
              "chartName": "string",
              "chartViews": 0,
              "lastUpdatedAt": "2019-08-24T14:15:22Z",
              "lastUpdatedBy": "string",
              "fieldName": "string",
              "chartType": "line",
              "chartUuid": "string"}]
        """
        resp = self._request("get", f"projects/{project_uuid}/validate")
        if resp.ok:
            return resp.json()["results"]
        return []

    def get_space(self, project_uuid: str, space_uuid: str):
        """
        Get details for a space in a project
        Return format:
            { "pinnedListOrder": 0,
              "pinnedListUuid": "string",
              "access": [ {
                  "role": "viewer",
                  "lastName": "string",
                  "firstName": "string",
                  "userUuid": "string" } ],
              "dashboards": [ {
                  "name": "string",
                  "organizationUuid": "string",
                  "uuid": "string",
                  "description": "string",
                  "updatedAt": "2019-08-24T14:15:22Z",
                  "projectUuid": "string",
                  "updatedByUser": {
                    "userUuid": "string",
                    "firstName": "string",
                    "lastName": "string" },
                  "spaceUuid": "string",
                  "views": 0,
                  "firstViewedAt": "2019-08-24T14:15:22Z",
                  "pinnedListUuid": "string",
                  "pinnedListOrder": 0,
                  "validationErrors": [ {
                      "validationId": 0,
                      "createdAt": "2019-08-24T14:15:22Z",
                      "error": "string" } ] } ],
              "projectUuid": "string",
              "queries": [ {
                  "name": "string",
                  "uuid": "string",
                  "description": "string",
                  "updatedAt": "2019-08-24T14:15:22Z",
                  "updatedByUser": {
                    "userUuid": "string",
                    "firstName": "string",
                    "lastName": "string" },
                  "spaceUuid": "string",
                  "pinnedListUuid": "string",
                  "pinnedListOrder": 0,
                  "firstViewedAt": "2019-08-24T14:15:22Z",
                  "views": 0,
                  "validationErrors": [ {
                      "validationId": 0,
                      "createdAt": "2019-08-24T14:15:22Z",
                      "error": "string" } ],
                  "chartType": "line" } ],
              "isPrivate": true,
              "name": "string",
              "uuid": "string",
              "organizationUuid": "string" }
        """
        resp = self._request("get", f"projects/{project_uuid}/spaces/{space_uuid}")
        if resp.ok:
            return resp.json()["results"]
        return {}

    def get_chart_version_history(self, chart_uuid: str):
        """
        Get chart version history from last 30 days
        Return format:
            [{"createdAt": "2019-08-24T14:15:22Z",
              "chartUuid": "string",
              "versionUuid": "string",
              "createdBy": {
                "userUuid": "string",
                "firstName": "string",
                "lastName": "string" }}]
        """
        resp = self._request("get", f"saved/{chart_uuid}/history")
        if resp.ok:
            return resp.json()["results"]["history"]
        return []

    def get_chart(self, chart_uuid: str, version_uuid: str):
        """
        Get chart details
        Return format:
            { "chart": {
                "dashboardName": "string",
                "dashboardUuid": "string",
                "pinnedListOrder": 0,
                "pinnedListUuid": "string",
                "spaceName": "string",
                "spaceUuid": "string",
                "organizationUuid": "string",
                "updatedByUser": {
                  "userUuid": "string",
                  "firstName": "string",
                  "lastName": "string" },
                "updatedAt": "2019-08-24T14:15:22Z",
                "tableConfig": {
                  "columnOrder": [
                    "string" ] },
                "chartConfig": {
                  "config": {
                    "legendPosition": "horizontal",
                    "showLegend": true,
                    "groupSortOverrides": [
                      "string" ],
                    "groupValueOptionOverrides": {},
                    "groupColorOverrides": {},
                    "groupLabelOverrides": {},
                    "showPercentage": true,
                    "showValue": true,
                    "valueLabel": "hidden",
                    "isDonut": true,
                    "metricId": "string",
                    "groupFieldIds": [
                      "string" ] },
                  "type": "pie" },
                "pivotConfig": {
                  "columns": [
                    "string" ] },
                "metricQuery": {
                  "additionalMetrics": [ {
                      "label": "string",
                      "type": "percentile",
                      "description": "string",
                      "sql": "string",
                      "hidden": true,
                      "round": 0,
                      "compact": "thousands",
                      "format": "km",
                      "table": "string",
                      "name": "string",
                      "index": 0,
                      "filters": [ {
                          "values": [
                            null ],
                          "operator": "isNull",
                          "id": "string",
                          "target": {
                            "fieldRef": "string" },
                          "settings": null,
                          "disabled": true } ],
                      "baseDimensionName": "string",
                      "uuid": "string",
                      "percentile": 0 } ],
                  "tableCalculations": [ {
                      "format": {
                        "suffix": "string",
                        "prefix": "string",
                        "compact": "thousands",
                        "currency": "string",
                        "separator": "default",
                        "round": 0,
                        "type": "default" },
                      "sql": "string",
                      "displayName": "string",
                      "name": "string",
                      "index": 0 } ],
                  "limit": 0,
                  "sorts": [ {
                      "descending": true,
                      "fieldId": "string" } ],
                  "filters": {
                    "metrics": {
                      "or": [
                        null ],
                      "id": "string" },
                    "dimensions": {
                      "or": [
                        null ],
                      "id": "string" } },
                  "metrics": [
                    "string" ],
                  "dimensions": [
                    "string" ] },
                "tableName": "string",
                "description": "string",
                "name": "string",
                "projectUuid": "string",
                "uuid": "string" },
              "createdBy": {
                "userUuid": "string",
                "firstName": "string",
                "lastName": "string" },
              "createdAt": "2019-08-24T14:15:22Z",
              "versionUuid": "string",
              "chartUuid": "string" }
        """
        resp = self._request("get", f"saved/{chart_uuid}/version/{version_uuid}")
        if resp.ok:
            return resp.json()["results"]
        return {}

    def get_scheduler_logs(self, project_uuid: str):
        """
        Get scheduled logs
        Return format:
            { "logs": [ {
                  "details": {},
                  "targetType": "email",
                  "target": "string",
                  "status": "scheduled",
                  "createdAt": "2019-08-24T14:15:22Z",
                  "scheduledTime": "2019-08-24T14:15:22Z",
                  "jobGroup": "string",
                  "jobId": "string",
                  "schedulerUuid": "string",
                  "task": "handleScheduledDelivery"
                } ],
              "dashboards": [ {
                  "dashboardUuid": "string",
                  "name": "string" } ],
              "charts": [ {
                  "savedChartUuid": "string",
                  "name": "string" } ],
              "users": [ {
                  "userUuid": "string",
                  "lastName": "string",
                  "firstName": "string" } ],
              "schedulers": [ {
                  "options": {
                    "limit": 0,
                    "formatted": true },
                  "dashboardUuid": null,
                  "savedChartUuid": "string",
                  "cron": "string",
                  "format": "csv",
                  "createdBy": "string",
                  "updatedAt": "2019-08-24T14:15:22Z",
                  "createdAt": "2019-08-24T14:15:22Z",
                  "message": "string",
                  "name": "string",
                  "schedulerUuid": "string",
                  "targets": [ {
                      "channel": "string",
                      "schedulerUuid": "string",
                      "updatedAt": "2019-08-24T14:15:22Z",
                      "createdAt": "2019-08-24T14:15:22Z",
                      "schedulerSlackTargetUuid": "string" } ] } ] }
        """
        resp = self._request("get", f"schedulers/{project_uuid}/logs")
        if resp.ok:
            return resp.json()["results"]
        return {}

    def get_scheduler(self, scheduler_uuid: str):
        """
        Get details of a scheduler
        Return format:
            { "options": {
                "limit": 0,
                "formatted": true },
              "dashboardUuid": null,
              "savedChartUuid": "string",
              "cron": "string",
              "format": "csv",
              "createdBy": "string",
              "updatedAt": "2019-08-24T14:15:22Z",
              "createdAt": "2019-08-24T14:15:22Z",
              "message": "string",
              "name": "string",
              "schedulerUuid": "string",
              "targets": [ {
                  "channel": "string",
                  "schedulerUuid": "string",
                  "updatedAt": "2019-08-24T14:15:22Z",
                  "createdAt": "2019-08-24T14:15:22Z",
                  "schedulerSlackTargetUuid": "string" } ] }
        """
        resp = self._request("get", f"schedulers/{scheduler_uuid}")
        if resp.ok:
            return resp.json()["results"]
        return {}

    def get_scheduler_jobs(self, scheduler_uuid: str):
        """
        Get jobs scheduled by a scheduler
        Return format:
            [ { "id": "string",
                "date": "2019-08-24T14:15:22Z" } ]
        """
        resp = self._request("get", f"schedulers/{scheduler_uuid}/jobs")
        if resp.ok:
            return resp.json()["results"]
        return []

    def get_scheduler_job_status(self, job_id: str):
        """
        Get a generic job status
        Return format:
            { "status": "string" }
        """
        resp = self._request("get", f"schedulers/job/{job_id}/status")
        if resp.ok:
            return resp.json()["results"]
        return {}



"""triggers

Revision ID: a57506731839
Revises: 607709e1615b
Create Date: 2023-07-13 18:24:59.186856

"""
from alembic import op
import sqlalchemy as sa
import mindsdb.interfaces.storage.db # noqa

revision = 'a57506731839'
down_revision = '607709e1615b'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        'tasks',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('company_id', sa.Integer(), nullable=True),
        sa.Column('user_class', sa.Integer(), nullable=True),
        sa.Column('object_type', sa.String(), nullable=False),
        sa.Column('object_id', sa.Integer(), nullable=False),
        sa.Column('last_error', sa.String(), nullable=True),
        sa.Column('active', sa.Boolean(), nullable=True),
        sa.Column('reload', sa.Boolean(), nullable=True),
        sa.Column('run_by', sa.String(), nullable=True),
        sa.Column('alive_time', sa.DateTime(timezone=True), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint('id')
    )
    op.create_table(
        'triggers',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('name', sa.String(), nullable=False),
        sa.Column('project_id', sa.Integer(), nullable=False),
        sa.Column('database_id', sa.Integer(), nullable=False),
        sa.Column('table_name', sa.String(), nullable=False),
        sa.Column('columns', sa.String(), nullable=True),
        sa.Column('query_str', sa.String(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint('id')
    )


def downgrade():
    op.drop_table('triggers')
    op.drop_table('tasks')
    # ### end Alembic commands ###
